@@ -71,6 +71,15 @@
 
 #include "gst_private.h"
 #include <gst/gst.h>
+#include <gobject/gvaluecollector.h>
+
+#ifdef GST_DISABLE_DEPRECATED
+#if !defined(GST_DISABLE_LOADSAVE) && !defined(GST_REMOVE_DEPRECATED)
+#include <libxml/parser.h>
+xmlNodePtr gst_caps_save_thyself (const GstCaps * caps, xmlNodePtr parent);
+GstCaps *gst_caps_load_thyself (xmlNodePtr parent);
+#endif
+#endif
 
 #define DEBUG_REFCOUNT
 
@@ -91,11 +100,27 @@
 #define IS_WRITABLE(caps) \
   (g_atomic_int_get (&(caps)->refcount) == 1)
 
+/* same as gst_caps_is_any () */
+#define CAPS_IS_ANY(caps)				\
+  ((caps)->flags & GST_CAPS_FLAGS_ANY)
+
+/* same as gst_caps_is_empty () */
+#define CAPS_IS_EMPTY(caps)				\
+  (!CAPS_IS_ANY(caps) && CAPS_IS_EMPTY_SIMPLE(caps))
+
+#define CAPS_IS_EMPTY_SIMPLE(caps)					\
+  (((caps)->structs == NULL) || ((caps)->structs->len == 0))
+
 /* quick way to get a caps structure at an index without doing a type or array
  * length check */
 #define gst_caps_get_structure_unchecked(caps, index) \
      ((GstStructure *)g_ptr_array_index ((caps)->structs, (index)))
-
+/* quick way to append a structure without checking the args */
+#define gst_caps_append_structure_unchecked(caps, structure) G_STMT_START{\
+  GstStructure *__s=structure;                                      \
+  gst_structure_set_parent_refcount (__s, &caps->refcount);         \
+  g_ptr_array_add (caps->structs, __s);                             \
+}G_STMT_END
 
 /* lock to protect multiple invocations of static caps to caps conversion */
 G_LOCK_DEFINE_STATIC (static_caps_lock);
@@ -200,7 +225,7 @@ gst_caps_new_simple (const char *media_type, const char *fieldname, ...)
   structure = gst_structure_new_valist (media_type, fieldname, var_args);
   va_end (var_args);
 
-  gst_caps_append_structure (caps, structure);
+  gst_caps_append_structure_unchecked (caps, structure);
 
   return caps;
 }
@@ -248,7 +273,7 @@ gst_caps_new_full_valist (GstStructure * structure, va_list var_args)
   caps = gst_caps_new_empty ();
 
   while (structure) {
-    gst_caps_append_structure (caps, structure);
+    gst_caps_append_structure_unchecked (caps, structure);
     structure = va_arg (var_args, GstStructure *);
   }
 
@@ -285,7 +310,8 @@ gst_caps_copy (const GstCaps * caps)
 
   for (i = 0; i < n; i++) {
     structure = gst_caps_get_structure_unchecked (caps, i);
-    gst_caps_append_structure (newcaps, gst_structure_copy (structure));
+    gst_caps_append_structure_unchecked (newcaps,
+        gst_structure_copy (structure));
   }
 
   return newcaps;
@@ -344,7 +370,7 @@ gst_caps_make_writable (GstCaps * caps)
   g_return_val_if_fail (caps != NULL, NULL);
 
   /* we are the only instance reffing this caps */
-  if (g_atomic_int_get (&caps->refcount) == 1)
+  if (IS_WRITABLE (caps))
     return caps;
 
   /* else copy */
@@ -365,8 +391,8 @@ gst_caps_make_writable (GstCaps * caps)
  * gst_caps_make_writable(), it is guaranteed that the caps object will not
  * change. This means its structures won't change, etc. To use a #GstCaps
  * object, you must always have a refcount on it -- either the one made
- * implicitly by gst_caps_new(), or via taking one explicitly with this
- * function.
+ * implicitly by e.g. gst_caps_new_simple(), or via taking one explicitly with
+ * this function.
  *
  * Returns: the same #GstCaps object.
  */
@@ -376,7 +402,7 @@ gst_caps_ref (GstCaps * caps)
   g_return_val_if_fail (caps != NULL, NULL);
 
 #ifdef DEBUG_REFCOUNT
-  GST_CAT_LOG (GST_CAT_REFCOUNTING, "%p %d->%d", caps,
+  GST_CAT_TRACE (GST_CAT_REFCOUNTING, "%p %d->%d", caps,
       GST_CAPS_REFCOUNT_VALUE (caps), GST_CAPS_REFCOUNT_VALUE (caps) + 1);
 #endif
   g_return_val_if_fail (GST_CAPS_REFCOUNT_VALUE (caps) > 0, NULL);
@@ -399,7 +425,7 @@ gst_caps_unref (GstCaps * caps)
   g_return_if_fail (caps != NULL);
 
 #ifdef DEBUG_REFCOUNT
-  GST_CAT_LOG (GST_CAT_REFCOUNTING, "%p %d->%d", caps,
+  GST_CAT_TRACE (GST_CAT_REFCOUNTING, "%p %d->%d", caps,
       GST_CAPS_REFCOUNT_VALUE (caps), GST_CAPS_REFCOUNT_VALUE (caps) - 1);
 #endif
 
@@ -500,6 +526,7 @@ no_string:
 }
 
 /* manipulation */
+
 static GstStructure *
 gst_caps_remove_and_get_structure (GstCaps * caps, guint idx)
 {
@@ -510,6 +537,30 @@ gst_caps_remove_and_get_structure (GstCaps * caps, guint idx)
   return s;
 }
 
+/**
+ * gst_caps_steal_structure:
+ * @caps: the #GstCaps to retrieve from
+ * @index: Index of the structure to retrieve
+ *
+ * Retrieves the stucture with the given index from the list of structures
+ * contained in @caps. The caller becomes the owner of the returned structure.
+ *
+ * Returns: a pointer to the #GstStructure corresponding to @index.
+ *
+ * Since: 0.10.30
+ */
+GstStructure *
+gst_caps_steal_structure (GstCaps * caps, guint index)
+{
+  g_return_val_if_fail (caps != NULL, NULL);
+  g_return_val_if_fail (IS_WRITABLE (caps), NULL);
+
+  if (G_UNLIKELY (index >= caps->structs->len))
+    return NULL;
+
+  return gst_caps_remove_and_get_structure (caps, index);
+}
+
 static gboolean
 gst_structure_is_equal_foreach (GQuark field_id, const GValue * val2,
     gpointer data)
@@ -609,7 +660,7 @@ gst_caps_append (GstCaps * caps1, GstCaps * caps2)
 #ifdef USE_POISONING
   CAPS_POISON (caps2);
 #endif
-  if (G_UNLIKELY (gst_caps_is_any (caps1) || gst_caps_is_any (caps2))) {
+  if (G_UNLIKELY (CAPS_IS_ANY (caps1) || CAPS_IS_ANY (caps2))) {
     /* FIXME: this leaks */
     caps1->flags |= GST_CAPS_FLAGS_ANY;
     for (i = caps2->structs->len - 1; i >= 0; i--) {
@@ -619,7 +670,7 @@ gst_caps_append (GstCaps * caps1, GstCaps * caps2)
   } else {
     for (i = caps2->structs->len; i; i--) {
       structure = gst_caps_remove_and_get_structure (caps2, 0);
-      gst_caps_append_structure (caps1, structure);
+      gst_caps_append_structure_unchecked (caps1, structure);
     }
   }
   gst_caps_unref (caps2);       /* guaranteed to free it */
@@ -651,12 +702,12 @@ gst_caps_merge (GstCaps * caps1, GstCaps * caps2)
 #ifdef USE_POISONING
   CAPS_POISON (caps2);
 #endif
-  if (G_UNLIKELY (gst_caps_is_any (caps1))) {
+  if (G_UNLIKELY (CAPS_IS_ANY (caps1))) {
     for (i = caps2->structs->len - 1; i >= 0; i--) {
       structure = gst_caps_remove_and_get_structure (caps2, i);
       gst_structure_free (structure);
     }
-  } else if (G_UNLIKELY (gst_caps_is_any (caps2))) {
+  } else if (G_UNLIKELY (CAPS_IS_ANY (caps2))) {
     caps1->flags |= GST_CAPS_FLAGS_ANY;
     for (i = caps1->structs->len - 1; i >= 0; i--) {
       structure = gst_caps_remove_and_get_structure (caps1, i);
@@ -701,8 +752,7 @@ gst_caps_append_structure (GstCaps * caps, GstStructure * structure)
     STRUCTURE_POISON (structure);
 #endif
 #endif
-    gst_structure_set_parent_refcount (structure, &caps->refcount);
-    g_ptr_array_add (caps->structs, structure);
+    gst_caps_append_structure_unchecked (caps, structure);
   }
 }
 
@@ -762,15 +812,13 @@ gst_caps_merge_structure (GstCaps * caps, GstStructure * structure)
       }
     }
     if (unique) {
-      gst_structure_set_parent_refcount (structure, &caps->refcount);
-      g_ptr_array_add (caps->structs, structure);
+      gst_caps_append_structure_unchecked (caps, structure);
     } else {
       gst_structure_free (structure);
     }
   }
 }
 
-
 /**
  * gst_caps_get_size:
  * @caps: a #GstCaps
@@ -802,7 +850,7 @@ gst_caps_get_size (const GstCaps * caps)
  * are writable, either because you have just copied them or made
  * them writable with gst_caps_make_writable(), you may modify the
  * structure returned in the usual way, e.g. with functions like
- * gst_structure_set_simple().
+ * gst_structure_set().
  *
  * You do not need to free or unref the structure returned, it
  * belongs to the #GstCaps.
@@ -841,7 +889,8 @@ gst_caps_copy_nth (const GstCaps * caps, guint nth)
 
   if (G_LIKELY (caps->structs->len > nth)) {
     structure = gst_caps_get_structure_unchecked (caps, nth);
-    gst_caps_append_structure (newcaps, gst_structure_copy (structure));
+    gst_caps_append_structure_unchecked (newcaps,
+        gst_structure_copy (structure));
   }
 
   return newcaps;
@@ -868,57 +917,111 @@ gst_caps_truncate (GstCaps * caps)
     gst_caps_remove_structure (caps, i--);
 }
 
+/**
+ * gst_caps_set_value:
+ * @caps: a writable caps
+ * @field: name of the field to set
+ * @value: value to set the field to
+ *
+ * Sets the given @field on all structures of @caps to the given @value.
+ * This is a convenience function for calling gst_structure_set_value() on
+ * all structures of @caps.
+ *
+ * Since: 0.10.26
+ **/
+void
+gst_caps_set_value (GstCaps * caps, const char *field, const GValue * value)
+{
+  guint i, len;
+
+  g_return_if_fail (GST_IS_CAPS (caps));
+  g_return_if_fail (IS_WRITABLE (caps));
+  g_return_if_fail (field != NULL);
+  g_return_if_fail (G_IS_VALUE (value));
+
+  len = caps->structs->len;
+  for (i = 0; i < len; i++) {
+    GstStructure *structure = gst_caps_get_structure_unchecked (caps, i);
+    gst_structure_set_value (structure, field, value);
+  }
+}
+
+/**
+ * gst_caps_set_simple_valist:
+ * @caps: the #GstCaps to set
+ * @field: first field to set
+ * @varargs: additional parameters
+ *
+ * Sets fields in a #GstCaps.  The arguments must be passed in the same
+ * manner as gst_structure_set(), and be NULL-terminated.
+ * <note>Prior to GStreamer version 0.10.26, this function failed when
+ * @caps was not simple. If your code needs to work with those versions 
+ * of GStreamer, you may only call this function when GST_CAPS_IS_SIMPLE()
+ * is %TRUE for @caps.</note>
+ */
+void
+gst_caps_set_simple_valist (GstCaps * caps, const char *field, va_list varargs)
+{
+  GValue value = { 0, };
+
+  g_return_if_fail (GST_IS_CAPS (caps));
+  g_return_if_fail (IS_WRITABLE (caps));
+
+  while (field) {
+    GType type;
+    char *err;
+
+    type = va_arg (varargs, GType);
+
+    if (G_UNLIKELY (type == G_TYPE_DATE)) {
+      g_warning ("Don't use G_TYPE_DATE, use GST_TYPE_DATE instead\n");
+      type = GST_TYPE_DATE;
+    }
+#if GLIB_CHECK_VERSION(2,23,3)
+    G_VALUE_COLLECT_INIT (&value, type, varargs, 0, &err);
+#else
+    g_value_init (&value, type);
+    G_VALUE_COLLECT (&value, varargs, 0, &err);
+#endif
+    if (G_UNLIKELY (err)) {
+      g_critical ("%s", err);
+      return;
+    }
+
+    gst_caps_set_value (caps, field, &value);
+
+    g_value_unset (&value);
+
+    field = va_arg (varargs, const gchar *);
+  }
+}
+
 /**
  * gst_caps_set_simple:
  * @caps: the #GstCaps to set
  * @field: first field to set
  * @...: additional parameters
  *
- * Sets fields in a simple #GstCaps.  A simple #GstCaps is one that
- * only has one structure.  The arguments must be passed in the same
+ * Sets fields in a #GstCaps.  The arguments must be passed in the same
  * manner as gst_structure_set(), and be NULL-terminated.
+ * <note>Prior to GStreamer version 0.10.26, this function failed when
+ * @caps was not simple. If your code needs to work with those versions
+ * of GStreamer, you may only call this function when GST_CAPS_IS_SIMPLE()
+ * is %TRUE for @caps.</note>
  */
 void
 gst_caps_set_simple (GstCaps * caps, const char *field, ...)
 {
-  GstStructure *structure;
   va_list var_args;
 
   g_return_if_fail (GST_IS_CAPS (caps));
-  g_return_if_fail (caps->structs->len == 1);
   g_return_if_fail (IS_WRITABLE (caps));
 
-  structure = gst_caps_get_structure_unchecked (caps, 0);
-
   va_start (var_args, field);
-  gst_structure_set_valist (structure, field, var_args);
+  gst_caps_set_simple_valist (caps, field, var_args);
   va_end (var_args);
 }
 
-/**
- * gst_caps_set_simple_valist:
- * @caps: the #GstCaps to copy
- * @field: first field to set
- * @varargs: additional parameters
- *
- * Sets fields in a simple #GstCaps.  A simple #GstCaps is one that
- * only has one structure.  The arguments must be passed in the same
- * manner as gst_structure_set(), and be NULL-terminated.
- */
-void
-gst_caps_set_simple_valist (GstCaps * caps, const char *field, va_list varargs)
-{
-  GstStructure *structure;
-
-  g_return_if_fail (GST_IS_CAPS (caps));
-  g_return_if_fail (caps->structs->len == 1);
-  g_return_if_fail (IS_WRITABLE (caps));
-
-  structure = gst_caps_get_structure_unchecked (caps, 0);
-
-  gst_structure_set_valist (structure, field, varargs);
-}
-
 /* tests */
 
 /**
@@ -934,7 +1037,7 @@ gst_caps_is_any (const GstCaps * caps)
 {
   g_return_val_if_fail (GST_IS_CAPS (caps), FALSE);
 
-  return (caps->flags & GST_CAPS_FLAGS_ANY);
+  return (CAPS_IS_ANY (caps));
 }
 
 /**
@@ -950,10 +1053,10 @@ gst_caps_is_empty (const GstCaps * caps)
 {
   g_return_val_if_fail (GST_IS_CAPS (caps), FALSE);
 
-  if (caps->flags & GST_CAPS_FLAGS_ANY)
+  if (CAPS_IS_ANY (caps))
     return FALSE;
 
-  return (caps->structs == NULL) || (caps->structs->len == 0);
+  return CAPS_IS_EMPTY_SIMPLE (caps);
 }
 
 static gboolean
@@ -1060,13 +1163,13 @@ gst_caps_is_subset (const GstCaps * subset, const GstCaps * superset)
   g_return_val_if_fail (subset != NULL, FALSE);
   g_return_val_if_fail (superset != NULL, FALSE);
 
-  if (gst_caps_is_empty (subset) || gst_caps_is_any (superset))
+  if (CAPS_IS_EMPTY (subset) || CAPS_IS_ANY (superset))
     return TRUE;
-  if (gst_caps_is_any (subset) || gst_caps_is_empty (superset))
+  if (CAPS_IS_ANY (subset) || CAPS_IS_EMPTY (superset))
     return FALSE;
 
   caps = gst_caps_subtract (subset, superset);
-  ret = gst_caps_is_empty (caps);
+  ret = CAPS_IS_EMPTY_SIMPLE (caps);
   gst_caps_unref (caps);
   return ret;
 }
@@ -1105,6 +1208,8 @@ gst_caps_is_equal (const GstCaps * caps1, const GstCaps * caps2)
   return gst_caps_is_subset (caps1, caps2) && gst_caps_is_subset (caps2, caps1);
 }
 
+/* intersect operation */
+
 typedef struct
 {
   GstStructure *dest;
@@ -1152,8 +1257,8 @@ gst_caps_structure_intersect (const GstStructure * struct1,
 {
   IntersectData data;
 
-  g_return_val_if_fail (struct1 != NULL, NULL);
-  g_return_val_if_fail (struct2 != NULL, NULL);
+  g_assert (struct1 != NULL);
+  g_assert (struct2 != NULL);
 
   if (G_UNLIKELY (struct1->name != struct2->name))
     return NULL;
@@ -1212,8 +1317,8 @@ static gboolean
 gst_caps_structure_can_intersect (const GstStructure * struct1,
     const GstStructure * struct2)
 {
-  g_return_val_if_fail (struct1 != NULL, FALSE);
-  g_return_val_if_fail (struct2 != NULL, FALSE);
+  g_assert (struct1 != NULL);
+  g_assert (struct2 != NULL);
 
   if (G_UNLIKELY (struct1->name != struct2->name))
     return FALSE;
@@ -1254,11 +1359,11 @@ gst_caps_can_intersect (const GstCaps * caps1, const GstCaps * caps2)
     return TRUE;
 
   /* empty caps on either side, return empty */
-  if (G_UNLIKELY (gst_caps_is_empty (caps1) || gst_caps_is_empty (caps2)))
+  if (G_UNLIKELY (CAPS_IS_EMPTY (caps1) || CAPS_IS_EMPTY (caps2)))
     return FALSE;
 
   /* one of the caps is any */
-  if (G_UNLIKELY (gst_caps_is_any (caps1) || gst_caps_is_any (caps2)))
+  if (G_UNLIKELY (CAPS_IS_ANY (caps1) || CAPS_IS_ANY (caps2)))
     return TRUE;
 
   /* run zigzag on top line then right line, this preserves the caps order
@@ -1305,48 +1410,6 @@ gst_caps_can_intersect (const GstCaps * caps1, const GstCaps * caps2)
   return FALSE;
 }
 
-#if 0
-static GstStructure *
-gst_caps_structure_union (const GstStructure * struct1,
-    const GstStructure * struct2)
-{
-  int i;
-  GstStructure *dest;
-  const GstStructureField *field1;
-  const GstStructureField *field2;
-  int ret;
-
-  /* FIXME this doesn't actually work */
-
-  if (struct1->name != struct2->name)
-    return NULL;
-
-  dest = gst_structure_id_empty_new (struct1->name);
-
-  for (i = 0; i < struct1->fields->len; i++) {
-    GValue dest_value = { 0 };
-
-    field1 = GST_STRUCTURE_FIELD (struct1, i);
-    field2 = gst_structure_id_get_field (struct2, field1->name);
-
-    if (field2 == NULL) {
-      continue;
-    } else {
-      if (gst_value_union (&dest_value, &field1->value, &field2->value)) {
-        gst_structure_set_value (dest, g_quark_to_string (field1->name),
-            &dest_value);
-      } else {
-        ret = gst_value_compare (&field1->value, &field2->value);
-      }
-    }
-  }
-
-  return dest;
-}
-#endif
-
-/* operations */
-
 /**
  * gst_caps_intersect:
  * @caps1: a #GstCaps to intersect
@@ -1376,13 +1439,13 @@ gst_caps_intersect (const GstCaps * caps1, const GstCaps * caps2)
     return gst_caps_copy (caps1);
 
   /* empty caps on either side, return empty */
-  if (G_UNLIKELY (gst_caps_is_empty (caps1) || gst_caps_is_empty (caps2)))
+  if (G_UNLIKELY (CAPS_IS_EMPTY (caps1) || CAPS_IS_EMPTY (caps2)))
     return gst_caps_new_empty ();
 
   /* one of the caps is any, just copy the other caps */
-  if (G_UNLIKELY (gst_caps_is_any (caps1)))
+  if (G_UNLIKELY (CAPS_IS_ANY (caps1)))
     return gst_caps_copy (caps2);
-  if (G_UNLIKELY (gst_caps_is_any (caps2)))
+  if (G_UNLIKELY (CAPS_IS_ANY (caps2)))
     return gst_caps_copy (caps1);
 
   dest = gst_caps_new_empty ();
@@ -1431,6 +1494,8 @@ gst_caps_intersect (const GstCaps * caps1, const GstCaps * caps2)
   return dest;
 }
 
+/* subtract operation */
+
 typedef struct
 {
   const GstStructure *subtract_from;
@@ -1438,7 +1503,6 @@ typedef struct
 }
 SubtractionEntry;
 
-
 static gboolean
 gst_caps_structure_subtract_field (GQuark field_id, const GValue * value,
     gpointer user_data)
@@ -1513,10 +1577,10 @@ gst_caps_subtract (const GstCaps * minuend, const GstCaps * subtrahend)
   g_return_val_if_fail (minuend != NULL, NULL);
   g_return_val_if_fail (subtrahend != NULL, NULL);
 
-  if (gst_caps_is_empty (minuend) || gst_caps_is_any (subtrahend)) {
+  if (CAPS_IS_EMPTY (minuend) || CAPS_IS_ANY (subtrahend)) {
     return gst_caps_new_empty ();
   }
-  if (gst_caps_is_empty (subtrahend))
+  if (CAPS_IS_EMPTY_SIMPLE (subtrahend))
     return gst_caps_copy (minuend);
 
   /* FIXME: Do we want this here or above?
@@ -1524,7 +1588,7 @@ gst_caps_subtract (const GstCaps * minuend, const GstCaps * subtrahend)
      ANY means for specific types, so it's not possible to reduce ANY partially
      You can only remove everything or nothing and that is done above.
      Note: there's a test that checks this behaviour. */
-  g_return_val_if_fail (!gst_caps_is_any (minuend), NULL);
+  g_return_val_if_fail (!CAPS_IS_ANY (minuend), NULL);
   sublen = subtrahend->structs->len;
   g_assert (sublen > 0);
 
@@ -1548,17 +1612,18 @@ gst_caps_subtract (const GstCaps * minuend, const GstCaps * subtrahend)
           GSList *walk;
 
           for (walk = list; walk; walk = g_slist_next (walk)) {
-            gst_caps_append_structure (dest, (GstStructure *) walk->data);
+            gst_caps_append_structure_unchecked (dest,
+                (GstStructure *) walk->data);
           }
           g_slist_free (list);
         } else {
-          gst_caps_append_structure (dest, gst_structure_copy (min));
+          gst_caps_append_structure_unchecked (dest, gst_structure_copy (min));
         }
       } else {
-        gst_caps_append_structure (dest, gst_structure_copy (min));
+        gst_caps_append_structure_unchecked (dest, gst_structure_copy (min));
       }
     }
-    if (gst_caps_is_empty (dest)) {
+    if (CAPS_IS_EMPTY_SIMPLE (dest)) {
       gst_caps_unref (src);
       return dest;
     }
@@ -1569,6 +1634,48 @@ gst_caps_subtract (const GstCaps * minuend, const GstCaps * subtrahend)
   return dest;
 }
 
+/* union operation */
+
+#if 0
+static GstStructure *
+gst_caps_structure_union (const GstStructure * struct1,
+    const GstStructure * struct2)
+{
+  int i;
+  GstStructure *dest;
+  const GstStructureField *field1;
+  const GstStructureField *field2;
+  int ret;
+
+  /* FIXME this doesn't actually work */
+
+  if (struct1->name != struct2->name)
+    return NULL;
+
+  dest = gst_structure_id_empty_new (struct1->name);
+
+  for (i = 0; i < struct1->fields->len; i++) {
+    GValue dest_value = { 0 };
+
+    field1 = GST_STRUCTURE_FIELD (struct1, i);
+    field2 = gst_structure_id_get_field (struct2, field1->name);
+
+    if (field2 == NULL) {
+      continue;
+    } else {
+      if (gst_value_union (&dest_value, &field1->value, &field2->value)) {
+        gst_structure_set_value (dest, g_quark_to_string (field1->name),
+            &dest_value);
+      } else {
+        ret = gst_value_compare (&field1->value, &field2->value);
+      }
+    }
+  }
+
+  return dest;
+}
+#endif
+
 /**
  * gst_caps_union:
  * @caps1: a #GstCaps to union
@@ -1589,13 +1696,13 @@ gst_caps_union (const GstCaps * caps1, const GstCaps * caps2)
   g_return_val_if_fail (caps1 != NULL, NULL);
   g_return_val_if_fail (caps2 != NULL, NULL);
 
-  if (gst_caps_is_empty (caps1))
+  if (CAPS_IS_EMPTY (caps1))
     return gst_caps_copy (caps2);
 
-  if (gst_caps_is_empty (caps2))
+  if (CAPS_IS_EMPTY (caps2))
     return gst_caps_copy (caps1);
 
-  if (gst_caps_is_any (caps1) || gst_caps_is_any (caps2))
+  if (CAPS_IS_ANY (caps1) || CAPS_IS_ANY (caps2))
     return gst_caps_new_any ();
 
   dest1 = gst_caps_copy (caps1);
@@ -1606,6 +1713,8 @@ gst_caps_union (const GstCaps * caps1, const GstCaps * caps2)
   return dest1;
 }
 
+/* normalize/simplify operations */
+
 typedef struct _NormalizeForeach
 {
   GstCaps *caps;
@@ -1627,7 +1736,7 @@ gst_caps_normalize_foreach (GQuark field_id, const GValue * value, gpointer ptr)
       GstStructure *structure = gst_structure_copy (nf->structure);
 
       gst_structure_id_set_value (structure, field_id, v);
-      gst_caps_append_structure (nf->caps, structure);
+      gst_caps_append_structure_unchecked (nf->caps, structure);
     }
 
     gst_value_init_and_copy (&val, gst_value_list_get_value (value, 0));
@@ -1654,15 +1763,14 @@ gst_caps_normalize (const GstCaps * caps)
 {
   NormalizeForeach nf;
   GstCaps *newcaps;
-  guint i, nlen;
+  guint i;
 
   g_return_val_if_fail (GST_IS_CAPS (caps), NULL);
 
   newcaps = gst_caps_copy (caps);
   nf.caps = newcaps;
-  nlen = newcaps->structs->len;
 
-  for (i = 0; i < nlen; i++) {
+  for (i = 0; i < gst_caps_get_size (newcaps); i++) {
     nf.structure = gst_caps_get_structure_unchecked (newcaps, i);
 
     while (!gst_structure_foreach (nf.structure,
@@ -1729,24 +1837,17 @@ gst_caps_structure_simplify (GstStructure ** result,
 
   /* try to subtract to get a real subset */
   if (gst_caps_structure_subtract (&list, simplify, compare)) {
-    switch (g_slist_length (list)) {
-      case 0:
-        *result = NULL;
-        return TRUE;
-      case 1:
-        *result = list->data;
-        g_slist_free (list);
-        return TRUE;
-      default:
-      {
-        GSList *walk;
-
-        for (walk = list; walk; walk = g_slist_next (walk)) {
-          gst_structure_free (walk->data);
-        }
-        g_slist_free (list);
-        break;
-      }
+    if (list == NULL) {         /* no result */
+      *result = NULL;
+      return TRUE;
+    } else if (list->next == NULL) {    /* one result */
+      *result = list->data;
+      g_slist_free (list);
+      return TRUE;
+    } else {                    /* multiple results */
+      g_slist_foreach (list, (GFunc) gst_structure_free, NULL);
+      g_slist_free (list);
+      list = NULL;
     }
   }
 
@@ -1858,7 +1959,9 @@ gst_caps_do_simplify (GstCaps * caps)
   return TRUE;
 }
 
-#ifndef GST_DISABLE_LOADSAVE
+/* persistence */
+
+#if !defined(GST_DISABLE_LOADSAVE) && !defined(GST_REMOVE_DEPRECATED)
 /**
  * gst_caps_save_thyself:
  * @caps: a #GstCaps structure
@@ -1920,7 +2023,7 @@ gst_caps_replace (GstCaps ** caps, GstCaps * newcaps)
 
   oldcaps = *caps;
 
-  GST_CAT_LOG (GST_CAT_REFCOUNTING, "%p, %p -> %p", caps, oldcaps, newcaps);
+  GST_CAT_TRACE (GST_CAT_REFCOUNTING, "%p, %p -> %p", caps, oldcaps, newcaps);
 
   if (newcaps != oldcaps) {
     if (newcaps)
@@ -1963,10 +2066,10 @@ gst_caps_to_string (const GstCaps * caps)
   if (caps == NULL) {
     return g_strdup ("NULL");
   }
-  if (gst_caps_is_any (caps)) {
+  if (CAPS_IS_ANY (caps)) {
     return g_strdup ("ANY");
   }
-  if (gst_caps_is_empty (caps)) {
+  if (CAPS_IS_EMPTY_SIMPLE (caps)) {
     return g_strdup ("EMPTY");
   }
 
@@ -2004,7 +2107,6 @@ gst_caps_from_string_inplace (GstCaps * caps, const gchar * string)
   GstStructure *structure;
   gchar *s;
 
-  g_return_val_if_fail (string, FALSE);
   if (strcmp ("ANY", string) == 0) {
     caps->flags = GST_CAPS_FLAGS_ANY;
     return TRUE;
@@ -2017,7 +2119,7 @@ gst_caps_from_string_inplace (GstCaps * caps, const gchar * string)
   if (structure == NULL) {
     return FALSE;
   }
-  gst_caps_append_structure (caps, structure);
+  gst_caps_append_structure_unchecked (caps, structure);
 
   do {
 
@@ -2030,7 +2132,7 @@ gst_caps_from_string_inplace (GstCaps * caps, const gchar * string)
     if (structure == NULL) {
       return FALSE;
     }
-    gst_caps_append_structure (caps, structure);
+    gst_caps_append_structure_unchecked (caps, structure);
 
   } while (TRUE);
 
@@ -2050,6 +2152,8 @@ gst_caps_from_string (const gchar * string)
 {
   GstCaps *caps;
 
+  g_return_val_if_fail (string, FALSE);
+
   caps = gst_caps_new_empty ();
   if (gst_caps_from_string_inplace (caps, string)) {
     return caps;
