@@ -258,6 +258,8 @@ struct _GstBaseTransformPrivate
   /* QoS stats */
   guint64 processed;
   guint64 dropped;
+
+  GstClockTime last_stop_out;
 };
 
 static GstElementClass *parent_class = NULL;
@@ -326,6 +328,8 @@ static gboolean gst_base_transform_acceptcaps_default (GstBaseTransform * trans,
 static gboolean gst_base_transform_setcaps (GstPad * pad, GstCaps * caps);
 static GstFlowReturn gst_base_transform_buffer_alloc (GstPad * pad,
     guint64 offset, guint size, GstCaps * caps, GstBuffer ** buf);
+static gboolean gst_base_transform_query (GstPad * pad, GstQuery * query);
+static const GstQueryType *gst_base_transform_query_type (GstPad * pad);
 
 /* static guint gst_base_transform_signals[LAST_SIGNAL] = { 0 }; */
 
@@ -402,6 +406,10 @@ gst_base_transform_init (GstBaseTransform * trans,
       GST_DEBUG_FUNCPTR (gst_base_transform_sink_activate_push));
   gst_pad_set_bufferalloc_function (trans->sinkpad,
       GST_DEBUG_FUNCPTR (gst_base_transform_buffer_alloc));
+  gst_pad_set_query_function (trans->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_base_transform_query));
+  gst_pad_set_query_type_function (trans->sinkpad,
+      GST_DEBUG_FUNCPTR (gst_base_transform_query_type));
   gst_element_add_pad (GST_ELEMENT (trans), trans->sinkpad);
 
   pad_template =
@@ -420,6 +428,10 @@ gst_base_transform_init (GstBaseTransform * trans,
       GST_DEBUG_FUNCPTR (gst_base_transform_getrange));
   gst_pad_set_activatepull_function (trans->srcpad,
       GST_DEBUG_FUNCPTR (gst_base_transform_src_activate_pull));
+  gst_pad_set_query_function (trans->srcpad,
+      GST_DEBUG_FUNCPTR (gst_base_transform_query));
+  gst_pad_set_query_type_function (trans->srcpad,
+      GST_DEBUG_FUNCPTR (gst_base_transform_query_type));
   gst_element_add_pad (GST_ELEMENT (trans), trans->srcpad);
 
   trans->transform_lock = g_mutex_new ();
@@ -711,7 +723,8 @@ gst_base_transform_configure_caps (GstBaseTransform * trans, GstCaps * in,
 
   GST_OBJECT_LOCK (trans);
   /* make sure we reevaluate how the buffer_alloc works wrt to proxy allocating
-   * the buffer. */
+   * the buffer. FIXME, this triggers some quite heavy codepaths that don't need
+   * to be taken.. */
   trans->priv->suggest_pending = TRUE;
   GST_OBJECT_UNLOCK (trans);
   trans->negotiated = ret;
@@ -1060,7 +1073,7 @@ done:
   /* ERRORS */
 no_transform_possible:
   {
-    GST_WARNING_OBJECT (trans,
+    GST_DEBUG_OBJECT (trans,
         "transform could not transform %" GST_PTR_FORMAT
         " in anything we support", caps);
     ret = FALSE;
@@ -1183,6 +1196,102 @@ failed_configure:
   }
 }
 
+static gboolean
+gst_base_transform_query (GstPad * pad, GstQuery * query)
+{
+  gboolean ret = FALSE;
+  GstBaseTransform *trans = GST_BASE_TRANSFORM (gst_pad_get_parent (pad));
+  GstPad *otherpad = (pad == trans->srcpad) ? trans->sinkpad : trans->srcpad;
+
+  switch (GST_QUERY_TYPE (query)) {
+    case GST_QUERY_POSITION:{
+      GstFormat format;
+
+      gst_query_parse_position (query, &format, NULL);
+      if (format == GST_FORMAT_TIME && trans->segment.format == GST_FORMAT_TIME) {
+        gint64 pos;
+        ret = TRUE;
+
+        if ((pad == trans->sinkpad)
+            || (trans->priv->last_stop_out == GST_CLOCK_TIME_NONE)) {
+          pos =
+              gst_segment_to_stream_time (&trans->segment, GST_FORMAT_TIME,
+              trans->segment.last_stop);
+        } else {
+          pos = gst_segment_to_stream_time (&trans->segment, GST_FORMAT_TIME,
+              trans->priv->last_stop_out);
+        }
+        gst_query_set_position (query, format, pos);
+      } else {
+        ret = gst_pad_peer_query (otherpad, query);
+      }
+      break;
+    }
+    default:
+      ret = gst_pad_peer_query (otherpad, query);
+      break;
+  }
+
+  gst_object_unref (trans);
+  return ret;
+}
+
+static const GstQueryType *
+gst_base_transform_query_type (GstPad * pad)
+{
+  static const GstQueryType types[] = {
+    GST_QUERY_POSITION,
+    GST_QUERY_NONE
+  };
+
+  return types;
+}
+
+static void
+compute_upstream_suggestion (GstBaseTransform * trans, guint expsize,
+    GstCaps * caps)
+{
+  GstCaps *othercaps;
+  GstBaseTransformPrivate *priv = trans->priv;
+
+  GST_DEBUG_OBJECT (trans, "trying to find upstream suggestion");
+
+  /* we cannot convert the current buffer but we might be able to suggest a
+   * new format upstream, try to find what the best format is. */
+  othercaps = gst_base_transform_find_transform (trans, trans->srcpad, caps);
+
+  if (!othercaps) {
+    GST_DEBUG_OBJECT (trans, "incompatible caps, ignoring");
+    /* we received caps that we cannot transform. Upstream is behaving badly
+     * because it should have checked if we could handle these caps. We can
+     * simply ignore these caps and produce a buffer with our original caps. */
+  } else {
+    guint size_suggest;
+
+    GST_DEBUG_OBJECT (trans, "getting size of suggestion");
+
+    /* not a subset, we have a new upstream suggestion, remember it and
+     * allocate a default buffer. First we try to convert the size */
+    if (gst_base_transform_transform_size (trans,
+            GST_PAD_SRC, caps, expsize, othercaps, &size_suggest)) {
+
+      /* ok, remember the suggestions now */
+      GST_DEBUG_OBJECT (trans,
+          "storing new caps and size suggestion of %u and %" GST_PTR_FORMAT,
+          size_suggest, othercaps);
+
+      GST_OBJECT_LOCK (trans->sinkpad);
+      if (priv->sink_suggest)
+        gst_caps_unref (priv->sink_suggest);
+      priv->sink_suggest = gst_caps_ref (othercaps);
+      priv->size_suggest = size_suggest;
+      trans->priv->suggest_pending = TRUE;
+      GST_OBJECT_UNLOCK (trans->sinkpad);
+    }
+    gst_caps_unref (othercaps);
+  }
+}
+
 /* Allocate a buffer using gst_pad_alloc_buffer
  *
  * This function can do renegotiation on the source pad
@@ -1289,6 +1398,7 @@ gst_base_transform_prepare_output_buffer (GstBaseTransform * trans,
   /* check if we got different caps on this new output buffer */
   newcaps = GST_BUFFER_CAPS (*out_buf);
   newsize = GST_BUFFER_SIZE (*out_buf);
+
   if (newcaps && !gst_caps_is_equal (newcaps, oldcaps)) {
     GstCaps *othercaps;
     gboolean can_convert;
@@ -1297,19 +1407,61 @@ gst_base_transform_prepare_output_buffer (GstBaseTransform * trans,
 
     incaps = GST_PAD_CAPS (trans->sinkpad);
 
+    /* check if we can convert the current incaps to the new target caps */
+    can_convert =
+        gst_base_transform_can_transform (trans, trans->sinkpad, incaps,
+        newcaps);
+
+    if (!can_convert) {
+      GST_DEBUG_OBJECT (trans, "cannot perform transform on current buffer");
+
+      gst_base_transform_transform_size (trans,
+          GST_PAD_SINK, incaps, GST_BUFFER_SIZE (in_buf), newcaps, &expsize);
+
+      compute_upstream_suggestion (trans, expsize, newcaps);
+
+      /* we got a suggested caps but we can't transform to it. See if there is
+       * another downstream format that we can transform to */
+      othercaps =
+          gst_base_transform_find_transform (trans, trans->sinkpad, incaps);
+
+      if (othercaps && !gst_caps_is_empty (othercaps)) {
+        GST_DEBUG_OBJECT (trans, "we found target caps %" GST_PTR_FORMAT,
+            othercaps);
+        *out_buf = gst_buffer_make_metadata_writable (*out_buf);
+        gst_buffer_set_caps (*out_buf, othercaps);
+        gst_caps_unref (othercaps);
+        newcaps = GST_BUFFER_CAPS (*out_buf);
+        can_convert = TRUE;
+      } else if (othercaps)
+        gst_caps_unref (othercaps);
+    }
+
     /* it's possible that the buffer we got is of the wrong size, get the
      * expected size here, we will check the size if we are going to use the
      * buffer later on. */
     gst_base_transform_transform_size (trans,
         GST_PAD_SINK, incaps, GST_BUFFER_SIZE (in_buf), newcaps, &expsize);
 
-    /* check if we can convert the current incaps to the new target caps */
-    can_convert =
-        gst_base_transform_can_transform (trans, trans->sinkpad, incaps,
-        newcaps);
-
     if (can_convert) {
       GST_DEBUG_OBJECT (trans, "reconfigure transform for current buffer");
+
+      /* subclass might want to add fields to the caps */
+      if (bclass->fixate_caps != NULL) {
+        newcaps = gst_caps_copy (newcaps);
+
+        GST_DEBUG_OBJECT (trans, "doing fixate %" GST_PTR_FORMAT
+            " using caps %" GST_PTR_FORMAT
+            " on pad %s:%s using fixate_caps vmethod", newcaps, incaps,
+            GST_DEBUG_PAD_NAME (trans->srcpad));
+        bclass->fixate_caps (trans, GST_PAD_SINK, incaps, newcaps);
+
+        *out_buf = gst_buffer_make_metadata_writable (*out_buf);
+        gst_buffer_set_caps (*out_buf, newcaps);
+        gst_caps_unref (newcaps);
+        newcaps = GST_BUFFER_CAPS (*out_buf);
+      }
+
       /* caps not empty, try to renegotiate to the new format */
       if (!gst_base_transform_configure_caps (trans, incaps, newcaps)) {
         /* not sure we need to fail hard here, we can simply continue our
@@ -1319,6 +1471,9 @@ gst_base_transform_prepare_output_buffer (GstBaseTransform * trans,
       /* new format configure, and use the new output buffer */
       gst_pad_set_caps (trans->srcpad, newcaps);
       discard = FALSE;
+      /* clear previous cached sink-pad caps, so buffer_alloc knows that
+       * it needs to revisit the decision about whether to proxy or not: */
+      gst_caps_replace (&priv->sink_alloc, NULL);
       /* if we got a buffer of the wrong size, discard it now and make sure we
        * allocate a propertly sized buffer later. */
       if (newsize != expsize) {
@@ -1328,43 +1483,8 @@ gst_base_transform_prepare_output_buffer (GstBaseTransform * trans,
       }
       outsize = expsize;
     } else {
-      GST_DEBUG_OBJECT (trans, "cannot perform transform on current buffer");
+      compute_upstream_suggestion (trans, expsize, newcaps);
 
-      /* we cannot convert the current buffer but we might be able to suggest a
-       * new format upstream, try to find what the best format is. */
-      othercaps =
-          gst_base_transform_find_transform (trans, trans->srcpad, newcaps);
-
-      if (!othercaps) {
-        GST_DEBUG_OBJECT (trans, "incompatible caps, ignoring");
-        /* we received caps that we cannot transform. Upstream is behaving badly
-         * because it should have checked if we could handle these caps. We can
-         * simply ignore these caps and produce a buffer with our original caps. */
-      } else {
-        guint size_suggest;
-
-        GST_DEBUG_OBJECT (trans, "getting size of suggestion");
-
-        /* not a subset, we have a new upstream suggestion, remember it and
-         * allocate a default buffer. First we try to convert the size */
-        if (gst_base_transform_transform_size (trans,
-                GST_PAD_SRC, newcaps, expsize, othercaps, &size_suggest)) {
-
-          /* ok, remember the suggestions now */
-          GST_DEBUG_OBJECT (trans,
-              "storing new caps and size suggestion of %u and %" GST_PTR_FORMAT,
-              size_suggest, othercaps);
-
-          GST_OBJECT_LOCK (trans->sinkpad);
-          if (priv->sink_suggest)
-            gst_caps_unref (priv->sink_suggest);
-          priv->sink_suggest = gst_caps_ref (othercaps);
-          priv->size_suggest = size_suggest;
-          trans->priv->suggest_pending = TRUE;
-          GST_OBJECT_UNLOCK (trans->sinkpad);
-        }
-        gst_caps_unref (othercaps);
-      }
       if (in_buf != *out_buf)
         gst_buffer_unref (*out_buf);
       *out_buf = NULL;
@@ -1550,6 +1670,7 @@ gst_base_transform_buffer_alloc (GstPad * pad, guint64 offset, guint size,
     GstCaps * caps, GstBuffer ** buf)
 {
   GstBaseTransform *trans;
+  GstBaseTransformClass *klass;
   GstBaseTransformPrivate *priv;
   GstFlowReturn res;
   gboolean proxy, suggest, same_caps;
@@ -1557,6 +1678,7 @@ gst_base_transform_buffer_alloc (GstPad * pad, guint64 offset, guint size,
   guint size_suggest;
 
   trans = GST_BASE_TRANSFORM (gst_pad_get_parent (pad));
+  klass = GST_BASE_TRANSFORM_GET_CLASS (trans);
   priv = trans->priv;
 
   GST_DEBUG_OBJECT (pad, "alloc with caps %p %" GST_PTR_FORMAT ", size %u",
@@ -1638,15 +1760,53 @@ gst_base_transform_buffer_alloc (GstPad * pad, guint64 offset, guint size,
           sink_suggest = NULL;
         }
 
-        GST_DEBUG_OBJECT (trans, "Caps fixated to now: %" GST_PTR_FORMAT,
+        GST_DEBUG_OBJECT (trans, "Caps fixed to: %" GST_PTR_FORMAT,
             sink_suggest);
       }
 
       if (sink_suggest) {
         templ = gst_pad_get_pad_template_caps (pad);
 
-        if (!gst_caps_can_intersect (sink_suggest, templ))
-          goto not_supported;
+        if (!gst_caps_can_intersect (sink_suggest, templ)) {
+          GstCaps *allowed;
+          GstCaps *peercaps;
+
+          GST_DEBUG_OBJECT (trans,
+              "Requested pad alloc caps are not supported: %" GST_PTR_FORMAT,
+              sink_suggest);
+          /* the requested pad alloc caps are not supported, so let's try
+           * picking something allowed between the pads (they are linked,
+           * there must be something) */
+          allowed = gst_pad_get_allowed_caps (pad);
+          if (allowed && !gst_caps_is_empty (allowed)) {
+            GST_DEBUG_OBJECT (trans,
+                "pads could agree on one of the following caps: " "%"
+                GST_PTR_FORMAT, allowed);
+            allowed = gst_caps_make_writable (allowed);
+
+            if (klass->fixate_caps) {
+              peercaps =
+                  gst_pad_get_allowed_caps (GST_BASE_TRANSFORM_SRC_PAD (trans));
+              klass->fixate_caps (trans, GST_PAD_SRC, peercaps, allowed);
+              gst_caps_unref (peercaps);
+            }
+
+            /* Fixate them to be safe if the subclass didn't do it */
+            gst_caps_truncate (allowed);
+            gst_pad_fixate_caps (pad, allowed);
+            gst_caps_replace (&sink_suggest, allowed);
+            gst_caps_unref (allowed);
+
+            suggest = TRUE;
+
+            GST_DEBUG_OBJECT (trans, "Fixated suggestion caps to %"
+                GST_PTR_FORMAT, sink_suggest);
+          } else {
+            if (allowed)
+              gst_caps_unref (allowed);
+            goto not_supported;
+          }
+        }
       }
     }
 
@@ -1688,7 +1848,8 @@ gst_base_transform_buffer_alloc (GstPad * pad, guint64 offset, guint size,
   GST_OBJECT_UNLOCK (pad);
 
   proxy = priv->proxy_alloc;
-  GST_DEBUG_OBJECT (trans, "doing default alloc, proxy %d", proxy);
+  GST_DEBUG_OBJECT (trans, "doing default alloc, proxy %d, suggest %d", proxy,
+      suggest);
 
   /* we only want to proxy if we have no suggestion pending, FIXME */
   if (proxy && !suggest) {
@@ -1720,8 +1881,11 @@ gst_base_transform_buffer_alloc (GstPad * pad, guint64 offset, guint size,
         GST_OBJECT_UNLOCK (pad);
       } else {
         GST_DEBUG_OBJECT (trans, "peer did not accept new caps");
-        /* peer does not accept the caps, free the buffer we received and
-         * create a buffer of the requested format by the default handler. */
+        /* peer does not accept the caps, disable proxy_alloc, free the
+         * buffer we received and create a buffer of the requested format
+         * by the default handler. */
+        GST_DEBUG_OBJECT (trans, "disabling proxy");
+        priv->proxy_alloc = FALSE;
         gst_buffer_unref (*buf);
         *buf = NULL;
       }
@@ -1810,6 +1974,7 @@ gst_base_transform_sink_eventfunc (GstBaseTransform * trans, GstEvent * event)
       /* we need new segment info after the flush. */
       trans->have_newsegment = FALSE;
       gst_segment_init (&trans->segment, GST_FORMAT_UNDEFINED);
+      trans->priv->last_stop_out = GST_CLOCK_TIME_NONE;
       break;
     case GST_EVENT_EOS:
       break;
@@ -1877,6 +2042,8 @@ gst_base_transform_src_eventfunc (GstBaseTransform * trans, GstEvent * event)
 {
   gboolean ret;
 
+  GST_DEBUG_OBJECT (trans, "handling event %p %" GST_PTR_FORMAT, event, event);
+
   switch (GST_EVENT_TYPE (event)) {
     case GST_EVENT_SEEK:
       break;
@@ -2042,10 +2209,17 @@ no_qos:
       GST_DEBUG_OBJECT (trans, "doing inplace transform");
 
       if (inbuf != *outbuf) {
-        /* different buffers, copy the input to the output first, we then do an
-         * in-place transform on the output buffer. */
-        memcpy (GST_BUFFER_DATA (*outbuf), GST_BUFFER_DATA (inbuf),
-            GST_BUFFER_SIZE (inbuf));
+        guint8 *indata, *outdata;
+
+        /* Different buffer. The data can still be the same when we are dealing
+         * with subbuffers of the same buffer. Note that because of the FIXME in
+         * prepare_output_buffer() we have decreased the refcounts of inbuf and
+         * outbuf to keep them writable */
+        indata = GST_BUFFER_DATA (inbuf);
+        outdata = GST_BUFFER_DATA (*outbuf);
+
+        if (indata != outdata)
+          memcpy (outdata, indata, GST_BUFFER_SIZE (inbuf));
       }
       ret = bclass->transform_ip (trans, *outbuf);
     } else {
@@ -2173,11 +2347,24 @@ gst_base_transform_chain (GstPad * pad, GstBuffer * buffer)
    * GST_BASE_TRANSFORM_FLOW_DROPPED we will not push either. */
   if (outbuf != NULL) {
     if ((ret == GST_FLOW_OK)) {
+      GstClockTime last_stop_out = GST_CLOCK_TIME_NONE;
+
       /* Remember last stop position */
-      if ((last_stop != GST_CLOCK_TIME_NONE) &&
-          (trans->segment.format == GST_FORMAT_TIME))
+      if (last_stop != GST_CLOCK_TIME_NONE &&
+          trans->segment.format == GST_FORMAT_TIME)
         gst_segment_set_last_stop (&trans->segment, GST_FORMAT_TIME, last_stop);
 
+      if (GST_BUFFER_TIMESTAMP_IS_VALID (outbuf)) {
+        last_stop_out = GST_BUFFER_TIMESTAMP (outbuf);
+        if (GST_BUFFER_DURATION_IS_VALID (outbuf))
+          last_stop_out += GST_BUFFER_DURATION (outbuf);
+      } else if (last_stop != GST_CLOCK_TIME_NONE) {
+        last_stop_out = last_stop;
+      }
+      if (last_stop_out != GST_CLOCK_TIME_NONE
+          && trans->segment.format == GST_FORMAT_TIME)
+        trans->priv->last_stop_out = last_stop_out;
+
       /* apply DISCONT flag if the buffer is not yet marked as such */
       if (trans->priv->discont) {
         if (!GST_BUFFER_IS_DISCONT (outbuf)) {
@@ -2263,6 +2450,7 @@ gst_base_transform_activate (GstBaseTransform * trans, gboolean active)
     trans->negotiated = FALSE;
     trans->have_newsegment = FALSE;
     gst_segment_init (&trans->segment, GST_FORMAT_UNDEFINED);
+    trans->priv->last_stop_out = GST_CLOCK_TIME_NONE;
     trans->priv->proportion = 1.0;
     trans->priv->earliest_time = -1;
     trans->priv->discont = FALSE;
