@@ -40,15 +40,29 @@
 #include "../../gst/gst-i18n-lib.h"
 
 #include <sys/types.h>
+
+#ifdef G_OS_WIN32
+#include <io.h>                 /* lseek, open, close, read */
+#undef lseek
+#define lseek _lseeki64
+#undef off_t
+#define off_t guint64
+#endif
+
 #include <sys/stat.h>
+#ifdef HAVE_SYS_SOCKET_H
 #include <sys/socket.h>
+#endif
 #include <fcntl.h>
 #include <stdio.h>
 #ifdef HAVE_UNISTD_H
 #include <unistd.h>
 #endif
 #ifdef _MSC_VER
-#include <io.h>
+#undef stat
+#define stat _stat
+#define fstat _fstat
+#define S_ISREG(m)	(((m)&S_IFREG)==S_IFREG)
 #endif
 #include <errno.h>
 #include <string.h>
@@ -111,6 +125,9 @@ static gboolean gst_fd_sink_start (GstBaseSink * basesink);
 static gboolean gst_fd_sink_stop (GstBaseSink * basesink);
 static gboolean gst_fd_sink_unlock (GstBaseSink * basesink);
 static gboolean gst_fd_sink_unlock_stop (GstBaseSink * basesink);
+static gboolean gst_fd_sink_event (GstBaseSink * sink, GstEvent * event);
+
+static gboolean gst_fd_sink_do_seek (GstFdSink * fdsink, guint64 new_offset);
 
 static void
 gst_fd_sink_base_init (gpointer g_class)
@@ -138,13 +155,12 @@ gst_fd_sink_class_init (GstFdSinkClass * klass)
   gobject_class->get_property = gst_fd_sink_get_property;
   gobject_class->dispose = gst_fd_sink_dispose;
 
-  gstbasesink_class->get_times = NULL;
   gstbasesink_class->render = GST_DEBUG_FUNCPTR (gst_fd_sink_render);
   gstbasesink_class->start = GST_DEBUG_FUNCPTR (gst_fd_sink_start);
   gstbasesink_class->stop = GST_DEBUG_FUNCPTR (gst_fd_sink_stop);
   gstbasesink_class->unlock = GST_DEBUG_FUNCPTR (gst_fd_sink_unlock);
   gstbasesink_class->unlock_stop = GST_DEBUG_FUNCPTR (gst_fd_sink_unlock_stop);
-  gstbasesink_class->event = NULL;
+  gstbasesink_class->event = GST_DEBUG_FUNCPTR (gst_fd_sink_event);
 
   g_object_class_install_property (gobject_class, ARG_FD,
       g_param_spec_int ("fd", "fd", "An open file descriptor to write to",
@@ -162,8 +178,9 @@ gst_fd_sink_init (GstFdSink * fdsink, GstFdSinkClass * klass)
   fdsink->fd = 1;
   fdsink->uri = g_strdup_printf ("fd://%d", fdsink->fd);
   fdsink->bytes_written = 0;
+  fdsink->current_pos = 0;
 
-  GST_BASE_SINK (fdsink)->sync = TRUE;
+  gst_base_sink_set_sync (GST_BASE_SINK (fdsink), FALSE);
 }
 
 static void
@@ -191,8 +208,7 @@ gst_fd_sink_query (GstPad * pad, GstQuery * query)
       switch (format) {
         case GST_FORMAT_DEFAULT:
         case GST_FORMAT_BYTES:
-          gst_query_set_position (query, GST_FORMAT_BYTES,
-              fdsink->bytes_written);
+          gst_query_set_position (query, GST_FORMAT_BYTES, fdsink->current_pos);
           return TRUE;
         default:
           return FALSE;
@@ -265,6 +281,7 @@ again:
   size -= written;
   data += written;
   fdsink->bytes_written += written;
+  fdsink->current_pos += written;
 
   GST_DEBUG_OBJECT (fdsink, "wrote %d bytes, %d left", written, size);
 
@@ -296,9 +313,9 @@ write_error:
         GST_ELEMENT_ERROR (fdsink, RESOURCE, NO_SPACE_LEFT, (NULL), (NULL));
         break;
       default:{
-        GST_ELEMENT_ERROR (fdsink, RESOURCE, WRITE,
-            (_("Error while writing to file descriptor \"%d\"."), fdsink->fd),
-            ("%s", g_strerror (errno)));
+        GST_ELEMENT_ERROR (fdsink, RESOURCE, WRITE, (NULL),
+            ("Error while writing to file descriptor %d: %s",
+                fdsink->fd, g_strerror (errno)));
       }
     }
     return GST_FLOW_ERROR;
@@ -330,20 +347,19 @@ gst_fd_sink_check_fd (GstFdSink * fdsink, int fd)
         goto not_seekable;
     }
   } else
-    GST_DEBUG_OBJECT (fdsink, "File descriptor \"%d\" is seekable", fd);
+    GST_DEBUG_OBJECT (fdsink, "File descriptor %d is seekable", fd);
 
   return TRUE;
 
 invalid:
   {
-    GST_ELEMENT_ERROR (fdsink, RESOURCE, WRITE,
-        (_("File descriptor \"%d\" is not valid."), fd),
-        ("%s", g_strerror (errno)));
+    GST_ELEMENT_ERROR (fdsink, RESOURCE, WRITE, (NULL),
+        ("File descriptor %d is not valid: %s", fd, g_strerror (errno)));
     return FALSE;
   }
 not_seekable:
   {
-    GST_DEBUG_OBJECT (fdsink, "File descriptor \"%d\" is a pipe", fd);
+    GST_DEBUG_OBJECT (fdsink, "File descriptor %d is a pipe", fd);
     return TRUE;
   }
 }
@@ -365,6 +381,9 @@ gst_fd_sink_start (GstBaseSink * basesink)
   gst_poll_add_fd (fdsink->fdset, &fd);
   gst_poll_fd_ctl_write (fdsink->fdset, &fd, TRUE);
 
+  fdsink->bytes_written = 0;
+  fdsink->current_pos = 0;
+
   return TRUE;
 
   /* ERRORS */
@@ -456,8 +475,6 @@ gst_fd_sink_set_property (GObject * object, guint prop_id,
 {
   GstFdSink *fdsink;
 
-  g_return_if_fail (GST_IS_FD_SINK (object));
-
   fdsink = GST_FD_SINK (object);
 
   switch (prop_id) {
@@ -469,6 +486,7 @@ gst_fd_sink_set_property (GObject * object, guint prop_id,
       break;
     }
     default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
 }
@@ -479,8 +497,6 @@ gst_fd_sink_get_property (GObject * object, guint prop_id, GValue * value,
 {
   GstFdSink *fdsink;
 
-  g_return_if_fail (GST_IS_FD_SINK (object));
-
   fdsink = GST_FD_SINK (object);
 
   switch (prop_id) {
@@ -488,10 +504,89 @@ gst_fd_sink_get_property (GObject * object, guint prop_id, GValue * value,
       g_value_set_int (value, fdsink->fd);
       break;
     default:
+      G_OBJECT_WARN_INVALID_PROPERTY_ID (object, prop_id, pspec);
       break;
   }
 }
 
+static gboolean
+gst_fd_sink_do_seek (GstFdSink * fdsink, guint64 new_offset)
+{
+  off_t result;
+
+  result = lseek (fdsink->fd, new_offset, SEEK_SET);
+
+  if (result == -1)
+    goto seek_failed;
+
+  fdsink->current_pos = new_offset;
+
+  GST_DEBUG_OBJECT (fdsink, "File desciptor %d to seek to position "
+      "%" G_GUINT64_FORMAT, fdsink->fd, fdsink->current_pos);
+
+  return TRUE;
+
+  /* ERRORS */
+seek_failed:
+  {
+    GST_DEBUG_OBJECT (fdsink, "File desciptor %d failed to seek to position "
+        "%" G_GUINT64_FORMAT, fdsink->fd, new_offset);
+    return FALSE;
+  }
+}
+
+static gboolean
+gst_fd_sink_event (GstBaseSink * sink, GstEvent * event)
+{
+  GstEventType type;
+  GstFdSink *fdsink;
+
+  fdsink = GST_FD_SINK (sink);
+
+  type = GST_EVENT_TYPE (event);
+
+  switch (type) {
+    case GST_EVENT_NEWSEGMENT:
+    {
+      gint64 start, stop, pos;
+      GstFormat format;
+      gst_event_parse_new_segment (event, NULL, NULL, &format, &start,
+          &stop, &pos);
+
+      if (format == GST_FORMAT_BYTES) {
+        /* only try to seek and fail when we are going to a different
+         * position */
+        if (fdsink->current_pos != start) {
+          /* FIXME, the seek should be performed on the pos field, start/stop are
+           * just boundaries for valid bytes offsets. We should also fill the file
+           * with zeroes if the new position extends the current EOF (sparse streams
+           * and segment accumulation). */
+          if (!gst_fd_sink_do_seek (fdsink, (guint64) start))
+            goto seek_failed;
+        }
+      } else {
+        GST_DEBUG_OBJECT (fdsink,
+            "Ignored NEWSEGMENT event of format %u (%s)", (guint) format,
+            gst_format_get_name (format));
+      }
+      break;
+    }
+    default:
+      break;
+  }
+
+  return TRUE;
+
+seek_failed:
+  {
+    GST_ELEMENT_ERROR (fdsink, RESOURCE, SEEK, (NULL),
+        ("Error while seeking on file descriptor %d: %s",
+            fdsink->fd, g_strerror (errno)));
+    return FALSE;
+  }
+
+}
+
 /*** GSTURIHANDLER INTERFACE *************************************************/
 
 static GstURIType
@@ -503,7 +598,7 @@ gst_fd_sink_uri_get_type (void)
 static gchar **
 gst_fd_sink_uri_get_protocols (void)
 {
-  static gchar *protocols[] = { "fd", NULL };
+  static gchar *protocols[] = { (char *) "fd", NULL };
 
   return protocols;
 }
